/*
    lkpatch - a Linux Kernel Patcher

    Copyright 1999, 2000 Theo Nolte <nolte@post.rwth-aachen.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <regex.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>

#include "lkpatch.h"

char*verstr(ver_t*v) { /* return version string */
 static char s[256]; static int pos=0;
 char*rv=s+pos;
  if(v->l[3]==INT_MAX) { /* no extraversion */
    sprintf(rv,"%d.%d.%d",v->l[0],v->l[1],v->l[2]);
  } else if(v->l[0]==2&&v->l[1]==3) { /* 2.3 pre */
    sprintf(rv,"%d.%d.%d-pre%d",v->l[0],v->l[1],v->l[2],v->l[3]);
  } else if(v->l[0]==2&&v->l[1]==4) { /* 2.4 test */
    sprintf(rv,"%d.%d.%d-test%d",v->l[0],v->l[1],v->l[2],v->l[3]);
  } else { /* unknown pre/test version */
    sprintf(rv,"unknown version %d.%d.%d-?%d",v->l[0],v->l[1],v->l[2],v->l[3]);
  }
  pos+=strlen(rv)+1;
  if(pos>sizeof(s)-30) pos=0;
  return(rv);
}

int strver(char*str,int l,ver_t*v) { /* cvt string to version */
 int rv=1;
 char*s=alloca(l+1);
  strncpy(s,str,l); s[l]='\0';
  if(strstr(s,"-pre")) { /* 2.3 pre version */
    if(sscanf(s,"%d.%d.%d-pre%d",&v->l[0],&v->l[1],&v->l[2],&v->l[3])!=4||
       v->l[0]!=2||v->l[1]!=3)
      rv=0;
  } else if(strstr(s,"-test")) { /* 2.4 test version */
    if(sscanf(s,"%d.%d.%d-test%d",&v->l[0],&v->l[1],&v->l[2],&v->l[3])!=4||
       v->l[0]!=2||v->l[1]!=4)
      rv=0;
  } else { /* no extraversion */
    v->l[3]=INT_MAX;
    if(sscanf(s,"%d.%d.%d",&v->l[0],&v->l[1],&v->l[2])!=3)
      rv=0;
  }
  return(rv);
}

void verchk(ver_t*v) { /* heuristic version nr check */
  if(v->l[0]<0||v->l[0]>10||
     v->l[1]<0||v->l[1]>10||
     v->l[2]<0||v->l[2]>200||
     v->l[3]<0||(v->l[3]>50&&v->l[3]!=INT_MAX)
    )
    error(1,0,"verchk: bad version %s",verstr(v));
}

int verlt(ver_t*v1,ver_t*v2) { /* v1 < v2 ? */
 int i;
  verchk(v1); verchk(v2);
  for(i=0;i<NVL;i++) {
   int d=v1->l[i]-v2->l[i];
    if(d<0) return(1);
    else if(d>0) return(0);
  }
  return(0);
}

int vereq(ver_t*v1,ver_t*v2) { /* v1 == v2 ? */
  verchk(v1); verchk(v2);
  return(!verlt(v1,v2)&&!verlt(v2,v1));
}

int verobs(ver_t*o,ver_t*n) { /* n obsolete */
 int i;
  verchk(o); verchk(n);
  if(verlt(o,n)) return(0);
  for(i=0;i<2;i++) {
    if(o->l[i]!=n->l[i]) error(1,0,"verobs: versions differ in major levels");
  }
  return(o->l[2]-n->l[2]>5);
}

void vercp(ver_t*d,ver_t*s) { /* copy s to d */
  verchk(s);
  memcpy(d,s,sizeof(ver_t));
}

void verzero(ver_t*v) { /* set version to zero */
 int i;
  for(i=0;i<NVL;i++) v->l[i]=0;
}

void getiver(char*kdir,ver_t*v) {
 char*f,s[128],s1[128],s2[128],s3[128],s4[128];
 FILE*fp;
 struct stat sbuf;
  verzero(v);
  f=emalloc(strlen(kdir)+strlen("/Makefile")+1);
  sprintf(f,"%s/Makefile",kdir);
  if(stat(f,&sbuf)) error(1,0,"%s does not seem to be kernel src dir",kdir);
  if(!(fp=fopen(f,"r"))) error(1,errno,"getiver: can't open %s",f);
  if(fscanf(fp,"VERSION = %s\n",s1)!=1) {
    error(1,errno,"getiver: can't read version");
  }
  if(fscanf(fp,"PATCHLEVEL = %s\n",s2)!=1) {
    error(1,errno,"getiver: can't read patchlevel");
  }
  if(fscanf(fp,"SUBLEVEL = %s\n",s3)!=1) {
    error(1,errno,"getiver: can't read sublevel");
  }
  if(fscanf(fp,"EXTRAVERSION = %s\n",s4)!=1) {
    error(1,errno,"getiver: can't read extraversion");
  }
  sprintf(s,"%s.%s.%s%s",s1,s2,s3,s4);
  if(!strver(s,strlen(s),v)) error(1,0,"getiver: bad version %s",s);
  if(fclose(fp)) error(1,errno,"can't close %s",f);
}

int inbuf(BUF*b,char*ndl) {
 char*hay=bufptr(b);
  dp(DALL,"inbuf: looking for %s",ndl);
  return(strstr(hay,ndl)!=0);
}

int verinbuf(ver_t*v,BUF*b) { /* is version v in buffer? */
  return(inbuf(b,verstr(v)));
}

int nextver(BUF*b,ver_t*cv,ver_t*lv,ver_t*nv) { /* guess next version */
  vercp(nv,cv);
  if(nv->l[3]!=INT_MAX) { /* try next pre-version */
    nv->l[3]++;
    if(verinbuf(nv,b)) return(1);
  }
  while(verlt(nv,lv)) { /* try next version */
    /* try preversion */
    nv->l[2]++; nv->l[3]=1;
    if(verinbuf(nv,b)) return(1);
    /* try regular */
    nv->l[3]=INT_MAX;
    if(verinbuf(nv,b)) return(1);
    /* must be gap */
  }
  return(0);
}

int getlv1(BUF*b,ver_t*v) { /* get latest version from LATEST-IS file */
 char*re="LATEST-IS-([0-9]+\\.[0-9]+\\.[0-9]+(-[a-z]+[0-9]+)?)[^A-Za-z0-9._-]";
 char*p;
 regex_t preg;
 regmatch_t match[5];
  dp(DTRC,"getlv1()");
  if(regcomp(&preg,re,REG_EXTENDED|REG_NEWLINE)) {
    error(1,0,"getlv1: regcomp failed");
  }
  verzero(v);
  for(p=bufptr(b);!regexec(&preg,p,2,match,0);p+=match[0].rm_eo) { /* match */
   ver_t v2;
    dp(DALL,"getlv1: found version string %.*s",
      match[1].rm_eo-match[1].rm_so,p+match[1].rm_so);
    if(!strver(p+match[1].rm_so,match[1].rm_eo-match[1].rm_so,&v2)) {
      error(1,0,"getlv1: can't strver latest version");
    }
    dp(DALL,"getlv1: found translated version: %s",verstr(&v2));
    if(verlt(v,&v2)) vercp(v,&v2);
    dp(DALL,"getlv1: latest so far: %s (v2:%s)",verstr(v),verstr(&v2));
  }
  if(!v->l[0]) return(-1);
  dp(DTRC,"getlv1: latest version is %s",verstr(v));
  return(0);
}

int getlv2(BUF*b,ver_t*v) { /* get latest version from files */
 char*p,*re[2];
 regex_t preg;
 regmatch_t match[10];
 int i;
  re[0]="linux-([0-9]+\\.[0-9]+\\.[0-9]+(-[a-z]+[0-9]+)?)\\.tar\\.(bz2|gz)";
  re[1]="patch-([0-9]+\\.[0-9]+\\.[0-9]+(-[a-z]+[0-9]+)?)\\.(bz2|gz)";
  verzero(v);
  for(i=0;i<2;i++) {
    if(regcomp(&preg,re[i],REG_EXTENDED|REG_NEWLINE)) {
      error(1,0,"getlv2: regcomp failed");
    }
    for(p=bufptr(b);!regexec(&preg,p,NEL(match),match,0);p+=match[0].rm_eo) {
     ver_t v2;
      if(!strver(p+match[1].rm_so,match[1].rm_eo-match[1].rm_so,&v2)) {
        error(1,0,"getlv2: can't sscanf version");
      }
 /*     dp(DALL,"getlv2: found translated version %s",verstr(&v2));*/
      if(verlt(v,&v2)) vercp(v,&v2);
/*      dp(DALL,"getlv2: latest so far: %s",verstr(v));*/
    }
  }
  if(!v->l[0]) return(-1);
  dp(DTRC,"getlv2: latest version is %s",verstr(v));
  return 0;
}

int getlv(BUF*ib,ver_t*ver) {
 int e1,e2;
 ver_t v;
  verzero(ver);
  if((e1=getlv1(ib,&v))>=0) { if(verlt(ver,&v)) vercp(ver,&v); }
  if((e2=getlv2(ib,&v))>=0) { if(verlt(ver,&v)) vercp(ver,&v); }
  return(e1>=0||e2>=0);
}

int getltv(BUF*ib,ver_t*ls,ver_t*lt) { /* get latest test version */
 int rv=0;
  
  return(rv);
}

static char*name(char*pre,ver_t*v,char*suf) {
 static char s[512];
  sprintf(s,"%s-%s.%s",pre,verstr(v),suf);
  dp(DALL,"name(%s,%s,%s): %s",pre,verstr(v),suf,s);
  return(s);
}

char*lkurl(char*lxurl,ver_t*v,char*suf) {
 char*s,*slash;
  s=emalloc(strlen(lxurl)+strlen("/kernel/vxxxx.yyyy/")+strlen(suf)+1);
  slash=lxurl[strlen(lxurl)-1]=='/'?"":"/";
  sprintf(s,"%s%skernel/v%d.%d/%s",lxurl,slash,v->l[0],v->l[1],suf);
  dp(DALL,"lkurl: \"%s\"",s);
  return(s);
}

static BUF*dlurl(UBUF*ub,int b,char*suf,int upd) {
 UBUFE*ue=ubel(ub,b);
 BUF*bb;
 int e;
 char*url=emalloc(strlen(ue->url)+strlen(suf)+1);
  sprintf(url,"%s%s",ue->url,suf);
  if((e=dlget(url,ue->vhost,bb=newbuf(),dlto))) {
    dp(DPRG,"download failed - giving this mirror a penalty");
    cfpen(ue->url,0.3);
    return(0);
  }
  if(upd) cfupd(ue->url,dlq());
  return(bb);
}

static BUF*dltool(UBUF*ub,int b,ver_t*v,char*suf,int upd) {
 char*p;
  p=alloca(20+strlen(suf));
  sprintf(p,"/kernel/v%d.%d/%s",v->l[0],v->l[1],suf);
  return(dlurl(ub,b,p,upd));
}

int apply(UBUF*ub,int b,BUF*ib,ver_t*iv,ver_t*nv) { /* apply patch for nv */
 char*s,*prog,*bzip,*gzip,cmd[256];
 BUF*bb;
  dp(DPRG,"%s: retrieving-",verstr(nv));
  gzip=progcache("gzip");
  bzip=progcache("bzip2");
  if(bzip&&inbuf(ib,s=name("patch",nv,"bz2"))) { /* get bzip-patch */
    prog="bzip2";
  } else if(gzip&&inbuf(ib,s=name("patch",nv,"gz"))) { /* get gzip-patch */
    prog="gzip";
  } else { /* no such patch */
    dp(DPRG,"No patch found for %s",verstr(nv));
/*    cfpen(ubel(ub,b)->url,0.2);*/
    ubdelel(ub,b);
    return(0);
  }
  if(!(bb=dltool(ub,b,nv,s,1))) return(0);
  dp(DPRG," - applying");
#if 0
  sprintf(cmd,"%s -cd %s| patch --dry-run -sf -d %s -p1",
    prog,buffile(bb),srcdir);
  if(system(cmd)) {
    error(1,0,"error: patch does not go in cleanly - source left unchanged");
  }
#endif
  sprintf(cmd,"%s -cd %s|sed 1q|grep -q %s",prog,buffile(bb),verstr(iv));
  dp(DALL,"executing \"%s\"",cmd);
  if(system(cmd)) error(1,0,"patch-%s not for %s",verstr(nv),verstr(iv));
  sprintf(cmd,"%s -cd %s| patch -sf -p1 -d %s",prog,buffile(bb),srcdir);
  dp(DALL,"executing \"%s\"",cmd);
  if(system(cmd)) error(1,0,"patch did not go in cleanly - stop");
  delbuf(bb);
  return(1);
}

int mirror(UBUF*ub,ver_t*cv,ver_t*lv) {
 int b,rv=0;
 BUF*ib;
  if((b=cfsel(ub))<0) error(1,0,"no mirror left to try");
  dp(DPRG,"selected mirror %s",ubel(ub,b)->url);
  if(!(ib=dltool(ub,b,cv,"",0))) {
    dp(DTRC,"index download failed");
    rv=-1; goto end2;
  }
  addbuf(ib,"\0",1);
  if(getlv(ib,lv)<0) {
    dp(DPRG,"mirror does not supply latest version");
    rv=-1; goto end1;
  }
  dp(DPRG,"mirror claims latest version is %s",verstr(lv));
  if(verlt(lv,cv)) {
    dp(DPRG,"mirror seems to be out of date");
    rv=-1; goto end1;
  }
  while(verlt(cv,lv)) {
   ver_t nv;
    if(!nextver(ib,cv,lv,&nv)) {
      dp(DPRG,"mirror inconsistent - can't find successor to %s",verstr(cv));
    }
    if(!apply(ub,b,ib,cv,&nv)) {
      error(0,0,"hmm.. error applying patch");
      rv=(-1); goto end1;
    }
    vercp(cv,&nv);
    rv=1;
  }
  end1: delbuf(ib);
  end2: ubdelel(ub,b);
  return(rv);
}

void patch(void) {
 ver_t iv,cv,lv,mlv;
 UBUF*ub;
  verzero(&lv);
  getiver(srcdir,&iv);
  vercp(&cv,&iv);
  dp(DPRG,"installed version is %s",verstr(&iv));
  ub=cfread();
  for(;;) {
   int e;
    if((e=mirror(ub,&cv,&mlv))<0) {
      dp(DTRC,"skipping this mirror");
      continue;
    }
    if(verlt(&lv,&mlv)) vercp(&lv,&mlv);
    if(e==1&&vereq(&cv,&lv)) break;
    else if(e==0&&--ntries<=0) break;
  }
  if(vereq(&iv,&cv)) {
    dp(DPRG,"sorry -- nothing new in the %d.%d series",iv.l[0],iv.l[1]);
  }
}

/*-------------------------- query --------------------------*/

char*rematch(char*s,char*re) {
 regex_t preq;
 regmatch_t match[10];
 char*rv=NULL,*p;
  regcomp(&preq,re,REG_EXTENDED|REG_NEWLINE);
  for(p=s;!regexec(&preq,p,NEL(match),match,0);p+=match[0].rm_eo) {
   char*m; int l;
    m=alloca(l=match[0].rm_eo-match[0].rm_so+1);
    strncpy(m,p+match[0].rm_so,l-1); m[l-1]='\0';
    if(!rv||strcmp(rv,m)<0) {
      if(rv) free(rv); rv=strdup(m);
    }
  }
  regfree(&preq);
  return(rv);
}


void query(void) {
 int try,b;
 ver_t iv,dv,sv,cv;
 char*ss,*ds,*ssuf,*dsuf;
 UBUF*ub=cfread();
 BUF*bb,*sb,*db,*ib;
  getiver(srcdir,&iv);
  printf("Installed version:          %s\n",verstr(&iv));
  for(try=0;ublen(ub)&&try<ntries;ubdelel(ub,b),try++) {
    if((b=cfsel(ub))<0) error(1,0,"no sites left to try in config file");
    dp(DPRG,"selected mirror %s",ubel(ub,b)->url);
    /* installed branch */
    if(!(ib=dltool(ub,b,&iv,"",0))) continue; addbuf(ib,"\0",1);
    getlv(ib,&cv);
    printf("Latest in installed branch: %s\n",verstr(&cv));
    if(!(bb=dlurl(ub,b,"/kernel/",0))) continue; addbuf(bb,"\0",1);
    /* stable branch */
    ss=rematch(bufptr(bb),"v[1-9]\\.[02468]");
    ssuf=alloca(strlen("/kernel/")+strlen(ss)+5);
    sprintf(ssuf,"/kernel/%s/",ss);
    if(!(sb=dlurl(ub,b,ssuf,0))) continue; addbuf(sb,"\0",1);
    getlv(sb,&sv);
    printf("Latest stable version:      %s\n",verstr(&sv));
    /* devel branch */
    ds=rematch(bufptr(bb),"v[1-9]\\.[13579]");
    dsuf=alloca(strlen("/kernel/")+strlen(ds)+5);
    sprintf(dsuf,"/kernel/%s/",ds);
    if(!(db=dlurl(ub,b,dsuf,0))) continue; addbuf(db,"\0",1);
    getlv(db,&dv);
    printf("Latest devel version:       %s\n",verstr(&dv));
    delbuf(bb); delbuf(sb); delbuf(db); delbuf(ib);
  }
}
