/*
    lkpatch - a Linux Kernel Patcher

    Copyright 1999, 2000 Theo Nolte <nolte@post.rwth-aachen.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "lkpatch.h"

char*dotquad(u32 ip) {
 static char s[50];
 int h=ntohl(ip),m=0xff;
  sprintf(s,"%d.%d.%d.%d",(h>>24)&m,(h>>16)&m,(h>>8)&m,h&m);
  return(s);
}

char*dotquad0(u32 ip) {
 static char s[50];
 int h=ntohl(ip),m=0xff;
  sprintf(s,"%03d.%03d.%03d.%03d",(h>>24)&m,(h>>16)&m,(h>>8)&m,h&m);
  return(s);
}

/*---------------------------DOWNLOAD-------------------------------*/

static long q;

long dlq(void) { return q; }

static int subst(char*s,char*pat,char*sub,char**p) {
 char*q; int rv;
  if((q=strstr(s,pat))) {
    *p=emalloc(strlen(s)+1-strlen(pat)+strlen(sub));
    sprintf(*p,"%.*s%s%s",q-s,s,sub,q+strlen(pat));
    rv=1;
  } else {
    *p=estrdup(s);
    rv=0;
  }
  return(rv);
}

char*dlcmd(char*url,char*vhost,char*of,int to) { /* assemble download cmd */
 char*cmdpat,*q1,*q2,hdr[512],sto[32];
  if(!url||!of) error(1,0,"dlcmd: null input");
  hdr[0]='\0';
  if(vhost) snprintf(hdr,sizeof(hdr),"Host: %s",vhost);
  snprintf(sto,sizeof(sto),"%d",to);
  if(dlpat) cmdpat=dlpat;
  else if(progpath("curl"))
    cmdpat="curl -Y 1 -y %to -s -H '%hdr' -o %ofile %url";
  else if(progpath("wget"))
    cmdpat="wget -T %to --header='%hdr' -q -O %ofile %url";
  else
    error(1,0,"need curl or wget for download");
  subst(cmdpat,"%url",url,&q1);
  subst(q1,"%ofile",of,&q2); free(q1);
  subst(q2,"%hdr",hdr,&q1); free(q2);
  subst(q1,"%to",sto,&q2); free(q1);
  return(q2);
}

int dlget(char*url,char*vhost,BUF*b,int to) {
 char*of,*cmd,buf[4096];
 int fd,e,n;
 long us;
 double bw;
 struct timeval t1,t2;
  of=tempnam("/tmp","lkp");
  cmd=dlcmd(url,vhost,of,to);
  dp(DDL,"dlget: dlcmd: '%s'",cmd);
  if(gettimeofday(&t1,NULL)==-1) error(1,errno,"dlget: gettimeofday1");
  if((e=system(cmd))) {
    free(cmd); unlink(of); return(e);
  }
  if(gettimeofday(&t2,NULL)==-1) error(1,errno,"dlget: gettimeofday2");
  if((fd=open(of,O_RDONLY))==-1) {
    error(1,errno,"dlget: can't open temp file %s",of);
  }
  while((n=read(fd,buf,sizeof(buf)))>0) {
    addbuf(b,buf,n);
  }
/*  mkfbuf(b,of);*/
  close(fd);
  us=1e6*(t2.tv_sec-t1.tv_sec)+(t2.tv_usec-t1.tv_usec);
  bw=1.e6*(double)buflen(b)/(double)us;
  q=(long)bw;
  free(cmd); unlink(of);
  return(0);
}

/*----------------------------------------------------------------------*/

struct progs_t {
  char*prog;
  char*path;
} progs[]={ {"sed"}, {"grep"} ,{"patch"}, {"gzip"}, {"bzip2"}, {0} };

char*progcache(char*s) {
 struct progs_t*pe;
  for(pe=progs;pe->prog;pe++) {
    if(streq(s,pe->prog)) return(pe->path);
  }
  return(0);
}

void checkenv(void) {
 struct progs_t *pe;
  for(pe=progs;pe->prog;pe++) {
    pe->path=progpath(pe->prog);
    if(streq(pe->prog,"gzip")||streq(pe->prog,"bzip2")) continue;
    if(!pe->path) error(1,0,"%s required",pe->prog);
  }
  if(!progcache("bzip2")&&!progcache("gzip")) {
    error(1,0,"gzip or bzip2 required");
  }
}
